#ifndef SIEVE_EXTPROGRAMS_LIMITS_H
#define SIEVE_EXTPROGRAMS_LIMITS_H

#define SIEVE_EXTPROGRAMS_MAX_PROGRAM_NAME_LEN 128
#define SIEVE_EXTPROGRAMS_MAX_PROGRAM_ARG_LEN  1024

#define SIEVE_EXTPROGRAMS_CONNECT_TIMEOUT_MSECS 5

#endif
